/*	===========================================================================
 UPC - EETAC - CSD - EMC - http://digsys.upc.edu

 * PRJ1 : Climate chamber  PHASE C: Digital control with hysteresis

 * Documentation location: https://digsys.upc.edu/csd/EED/chamber/chamber.html
 * Hardware for prototyping: Arduino Uno board. Chip ATmega328P
 * Arduino project code inherited from Phase B
				
The hardware is adapted to the requirements of the climate chamber 

Hardware --> The same as in Phase B. We simple rename SW_CW to be SW_HYS
			
	 --> A switch SW_HYS(hysteresis) is used for enabling hysteresis control
			- SW_HYS = '0' -->  ON/OFF hysteresis window control 
			- SW_HYS = '1' -->  Linear PWM control in the window

Software --> We will rename "ON_OFF_CONTROL" from phase B to be "HYS_CONTROL"
	--> Execute the output_logic() in this state 
		only when new temperature values are acquired. 
		This means a new indicator var_NM_flag that will be one when a 
		different temperature VR or VS is acquired. 
		
	--> ON/FF hysteresis Perform a hysteresis control 
	Zone 1: D = 100% (ON) while TS < TH --> full power applied to the load
	Zone 2: D = 0% (OFF) while TS > TL  --> zero power applied to the load 


	--> Linear PWM: Perform a hysteresis control in the three zones defined in the 
		algorithm: 
		Zone 1: D = 100% (ON) while TS < TL --> full power applied to the load
		Zone 2: Linear PWM modulation while  TL < TS < TH  --> a % of full power
		Zone 3: D = 0% (OFF) while TS > TL --> zero power applied to the load 
				

=============================================================================*/
/*	===========================================================================
	Include section
	=========================================================================*/
/*-----( Import needed libraries )-----*/
// <Wire.h> already pre-installed in Arduino IDE         
#include <Wire.h>      
#include <LiquidCrystal_PCF8574.h>

 // set the LCD address to 0x27 for a 16 chars and 2 line display
LiquidCrystal_PCF8574 lcd(0x27); 


/*	===========================================================================
	Definitions
	======================================================================== */

// Switches 
const byte pin_SENSOR		= A1;
const byte pin_REFERENCE	= A2;


// Start-stop push-button. Pin PD3 is also the microcontroller's external INT1
const byte INT1_pin_ST_BUTTON = 3;  // 

// Hysteresis switch is IO4
// When 1 we use the phase C algorithm 
const byte pin_SW_HYS = 4;

// LED for generating the 1 Hz squared waveform when acquiring (LED blinking)
const byte pin_SLED = 8; 

// Control signal output (also it will be the PWM VC for analogue write)
const byte pin_Vc = 9; 


// FSM list of states 
const char IDLE				= 'A'; 
const char SET_ACQ_TIMER	= 'B';
const char Do_NOTHING		= 'C';
const char ACQUIRE			= 'D';
const char TEMP_MEAS		= 'E';
const char HYS_CONTROL		= 'F';
const char TRANSMIT			= 'G';
const char STOP_ACQ_TIMER	= 'H';


// Default timer period to sample analogue inputs 
const unsigned long SAMPLING_PERIOD = 499; // 500 ms   (2 Hz sampling frequency)


const int HYS_WINDOW = 1.5; // 1.5 C
const int D_MAX = 99; //full power
const int D_MIN = 1;  // zero power

/*	===========================================================================
	Global RAM variables
	======================================================================== */
byte   	var_SW_HYS;	// Select hysteresis algorithm   
byte   	var_SLED; 	// Squared waveform when running

char 	var_current_state ;    // This is the key variable: FSM state memory

// sampling period variable, in case of adjusting the sampling frequency
static unsigned int var_SP; 

static unsigned int i;  // measurement array index to save current acquisitions
static unsigned int p;  // P = 0, 1 for printing to the LCD 

// Voltage variables are recorded in a circular string
static unsigned int var_Vr[10];   
static unsigned int var_Vs[10]; 

// Current and previous mean values
static unsigned int var_Vs_m;
static unsigned int var_Vr_m;

static unsigned int var_Vs_m_previous;
static unsigned int var_Vr_m_previous;

// Temperature variables will be calculated as a mean value
static float var_Ts; 
static float var_Tr;   

// variables for performing hysteresis control
static unsigned int var_VH;
static unsigned int var_VL;
static unsigned int var_V_hys; 
static byte var_Zone;  
static unsigned int var_D;
static unsigned int var_PWM;

// For timing the square waveform: 
static unsigned long var_previous_time;   
static unsigned long var_current_time;    // current time in ms
static unsigned long var_time_elapsed;   //  time elapsed since last flag


// Sample period flag will be set after the timing period (Ts) time has elapsed
byte 	var_SP_flag; 

// Start/stop flag will be set when detecting an interrupt from the push-button
byte 	var_ST_flag; 

// The system will poll internal timer when acquiring
byte   	var_TMR_flag;

// This indicator is set only when there is new information to write to LCD
byte   	var_LCD_flag;

// This indicator detects that a different Vs or Vr is acquired, and so, 
// the new Vc value has to ve calculated. 
byte   	var_NM_flag;  //(new measurement)


/*	===========================================================================
	System initialisation
	=========================================================================*/
void setup() { 
  int error;

// Output where to connect the sampling LED that will toggle every Ts	
	pinMode(pin_SLED, OUTPUT);	
	
// Control output where to connect the power driver and the pilot LED
	pinMode(pin_Vc, OUTPUT);	

// Configuring digital input pin for reading the desired control wave 
	pinMode(pin_SW_HYS, INPUT);  

// This is the special pin to be an external interrupt source INT1:
	pinMode(INT1_pin_ST_BUTTON, INPUT_PULLUP); 
	
// Serial monitor initialisation
	Serial.begin(9600); 
 
//Initialise the FSM
	var_current_state = IDLE;
	var_SLED = 0;  //No blinking
	//var_Vc 	 = 0;  //No power to the load

	var_D = D_MIN;
	var_Zone = 1;	// Let us imagine the chamber cool with Ts < TH
	
// Keep the timer off until it is required when running acquiring temperature
	var_TMR_flag = 0;
	var_current_time = 0; 
	var_previous_time = 0; 

		// Serial monitor ASCII message:
	Serial.println("Project PRJ1 - EMC");
	Serial.println("Climate chamber phase C: Hysteresis control"); 
	Serial.println("System IDLE: waiting for ST click to start"); 


//	---------------------------------------------------------------------------
// This is the interrupt handler. Read the Arduino documentation at: 
// https://www.arduino.cc/reference/en/language/functions/external-interrupts/attachinterrupt/
// attachInterrupt(digitalPinToInterrupt(pin), ISR, mode) (recommended)
	
	attachInterrupt(digitalPinToInterrupt(INT1_pin_ST_BUTTON), interrupt_service_routine_INT1, FALLING);
	// Interrupt when detecting a falling edge to start and stop operations

//	---------------------------------------------------------------------------
// See http://playground.arduino.cc/Main/I2cScanner how to check 
// the detection of the I2C device.
  Wire.begin();
  Wire.beginTransmission(0x27);
  error = Wire.endTransmission();
  Serial.print("Error: ");
  Serial.print(error);

  if (error == 0) {
    Serial.println(": LCD found correctly.");
   
   // Initialise the LCD using LCD for 16 chars 2 lines, turn on backlight
    lcd.begin(16, 2);  // initialize the LCD

  } else {
	// in this way we can verify the correctness of the I2C wiring SDA and SCL
    Serial.println(": LCD not found.");
  }

// First message
    lcd.setBacklight(255);
    lcd.home();
    lcd.clear();
    lcd.print("Hello LCD");
    delay(400);

    lcd.setBacklight(0);
    delay(400);
    lcd.setBacklight(255);

//-------- Write characters on the display ------------------
// NOTE: Cursor Position: (CHAR, LINE) 
	lcd.setCursor(0,0); //Start at character 0 on line 0
	lcd.print("Chamber Control ");
	delay(500);
	lcd.setCursor(0,1);//Start at character 0 on line 1
	lcd.print("Phase C: HYS-WIN");
	delay(2000);  // 2 seconds

// Clean screen
	lcd.clear();

//  Allow new LCD messages in the main loop
	var_LCD_flag = 1; 

	p = 0; // initialise the index for printing on the LCD

// Hysteresis equivalent voltage:
var_V_hys = HYS_WINDOW*10.0823;

}

/*	===========================================================================
	Infinite loop 
	======================================================================== */
void loop() {       
	read_inputs(); 		// capture the switch voltage values into RAM variables
	state_logic();    // calculating the next state to go
	output_logic();		// calculating outputs 
	write_outputs(); // Convert RAM variable values into pin output voltage
}

/*	===========================================================================
	State logic to determine where to go next
	======================================================================== */
void state_logic(void){

	update_time_flag();  // To track real time and be able to 
						 // acquire signals at the specified sampling frequency

	switch (var_current_state) {
	case IDLE:	// Do nothing, wait for start signal
		if (var_ST_flag == 1) {	  // start the process
			  var_current_state = SET_ACQ_TIMER;
			  var_ST_flag = 0;	  //Clear flag when one used
		}
		  else
			var_current_state = IDLE;     
		break;

	case SET_ACQ_TIMER:	// Do nothing, wait for start signal
		var_current_state = Do_NOTHING;       
		break;

	case Do_NOTHING:	// Wait for sampling period
		if (var_SP_flag == 1) {	  // start acquiring 
			var_current_state = ACQUIRE;
			var_SP_flag = 0;	  //Clear flag when one used
		}
		  else
			var_current_state = Do_NOTHING;      
		break;

	case ACQUIRE:
		var_current_state = TEMP_MEAS;     
		break;

	case TEMP_MEAS:	
		var_current_state = HYS_CONTROL;  
		break;
		
	case HYS_CONTROL:	
		var_current_state = TRANSMIT;  
		break;

	case TRANSMIT:
		if (var_ST_flag == 1) {	  // stop acquiring 
			var_current_state = STOP_ACQ_TIMER;
			var_ST_flag = 0;	  //Clear flag after being used used
		}
		  else
			var_current_state = Do_NOTHING;      
		break;
		
	case STOP_ACQ_TIMER:	// go to Idle to wait for another Start pulse
		var_current_state = IDLE;     
		break;

	}	
}

/*	===========================================================================
	Output logic 
	======================================================================== */
void output_logic(void){
int var_buff, j;
int var_dif;
	
	switch (var_current_state) {
	case IDLE:
		var_SLED = 0;  //No blinking
		// var_Vc 	 = 0;  //No power to the load
		var_D = D_MIN;
		var_Zone = 1;	// When the control is stopped, 
		var_PWM = 0;

		if (var_LCD_flag == 1){
			lcd.setCursor(0,0); //Start at character 4 on line 0
			lcd.print("..... IDLE .....");
			lcd.setCursor(0,1);
			lcd.print("Click ST button.");

		// Serial monitor ASCII message:			
			Serial.println("System IDLE: waiting for ST click to start"); 		
			
			var_LCD_flag = 0; // Allows new information in main loop		
		}

		break;

	case SET_ACQ_TIMER:	
		var_SLED = 0;  //No blinking
		var_SP = SAMPLING_PERIOD; 		
		
		for (i = 0; i <10; i++){  // initialise voltage buffer
			var_Vs[i] = 0;	var_Vr[i] = 0;	
		}
		i = 0; 
		var_Vs_m_previous = 0;
		var_Vr_m_previous = 0;

		// Turn on the timer for counting sampling periods
		var_TMR_flag = 1;
		var_current_time = 0; 
		var_previous_time = 0; 	 

		lcd.setCursor(0,0); //Start at character 4 on line 0
		lcd.print(" ON/OFF control ");
		break;

	case Do_NOTHING: // waiting for sampling period ...
		var_D = var_D;	

		break;	


	case ACQUIRE: 
		// Toggle operation with a XOR simply to indicate running acquisition
		// bitwise XOR operation 
		var_SLED = var_SLED^0b00000001; 
		
			// Two consecutive analogue to digital conversions
		var_Vs[i] = analogRead(pin_SENSOR);// read V_AD1 voltage (sensor temp.)
		var_Vr[i] = analogRead(pin_REFERENCE);// read V_AD2 voltage (reference temp.)

			// using mean values
		var_Vs_m = 0; 
		var_Vr_m = 0; 
		
		for (j = 0; j < 10; j++){
			var_Vs_m = var_Vs[j] + var_Vs_m;
			var_Vr_m = var_Vr[j] + var_Vr_m;
		}
		var_Vs_m = var_Vs_m/10;
		var_Vr_m = var_Vr_m/10;
		
		if ((var_Vs_m != var_Vs_m_previous) || (var_Vr_m != var_Vr_m_previous) ) 
			var_NM_flag = 1; //calculate temperatures and control
		
		else 
			var_NM_flag = 0; 
		
		var_Vs_m_previous = var_Vs_m; 
		var_Vr_m_previous = var_Vr_m;		
		
		i++ ; // circular buffer
		if (i == 10) i = 0;

		break;	

	case TEMP_MEAS:	// measurement equation

		if (var_NM_flag ==1){
			// Temperatures [ Vs = 49.23 mV/C  Ts(C) + 1.146 V ] 
								//	[var_Vs = (1024/5)Vs]
			var_Ts = (var_Vs_m - 234.701)/10.082313; 
			var_Tr = (var_Vr_m - 234.701)/10.082313; 

		// Rounding a float with only one decimal digit
			var_Ts = var_Ts*10 + 0.5;
			var_buff = (int)var_Ts;
			var_Ts = (float)var_buff/10;

			var_Tr = var_Tr*10 + 0.5;
			var_buff = (int)var_Tr;
			var_Tr = (float)var_buff/10;
		
		}
		break;

	case HYS_CONTROL: 
//	---------------------------------------------------------------------------
				// control equation processing only when there are new values 
		if (var_NM_flag == 1){
				//Voltages that define the hysteresis window
			var_VL = var_Vr_m - var_V_hys/2; 
			var_VH = var_Vr_m + var_V_hys/2; 
			
			if (var_SW_HYS == 0){  // ON/OFF hysteresis in two zones 	
				if (var_Zone == 1) {		
					if (var_Vs_m >  var_VH){
						var_D = D_MIN;	
						var_Zone = 2;  //change working zone
					}
					else 	
						var_D = D_MAX;  //keep the same zone
				}
				else {  // this is Zone 2
					if (var_Vs_m <  var_VL){
						var_D = D_MAX;	
						var_Zone = 1;  //change working zone
					}
					else 	
						var_D = D_MIN; //keep the same zone	
				}
			}
//	---------------------------------------------------------------------------
			else {	// Linear PWM hysteresis control in three zones

				switch (var_Zone) {
				case 1: // Zone 1: Max power. Duty cycle 100%
					if (var_Vs_m >  var_VL)
						var_Zone = 2;
					var_D = D_MAX;						
					break;
				
				case 2: // linear control of the duty cycle when heating
					if (var_Vs_m >  var_VH)
						var_Zone = 3;
					var_dif = var_Vr_m - var_Vs_m;
					if (var_dif > 0) // negative numbers 
						var_D = 50.0 + 6.48*(var_Vr_m - var_Vs_m);
					else 
						var_D = 50.0 - 6.48*(var_Vs_m - var_Vr_m);
					break;

				case 3: // Zone 3: Min power. Duty cycle 0%
					if (var_Vs_m <  var_VL)
						var_Zone = 1;
					var_D = D_MIN;	
					break;
				}
			}
		var_PWM = 2.55*var_D;
		}
		break;

	case TRANSMIT:	// measurement equation	
	
		// Serial monitor:
		Serial.print("Ts = "); 	
		Serial.print((float)var_Ts, 1); 		
		Serial.println(" degC     "); 

		Serial.print("Tr = "); 	
		Serial.print((float)var_Tr, 1); 		
		Serial.println(" degC     "); 


		Serial.print("D = "); 	
		Serial.print((float)var_D, 1); 		
		Serial.println(" %        "); 

//	---------------------------------------------------------------------------
// We can print to the LCD one temperature every second
		if (var_LCD_flag == 1){
			switch (p) {
				case 0:
					lcd.setCursor(0,1); //Start at character 0 on line 1
					lcd.print("Temp Ts =   ");
					lcd.setCursor(12,1); //Start at character 12 on line 1
					lcd.print((float)var_Ts);	
					break;
				case 1:
					lcd.setCursor(0,1); //Start at character 0 on line 1
					lcd.print("Temp Tr =   ");
					lcd.setCursor(12,1); //Start at character 12 on line 1
					lcd.print((float)var_Tr);	
					break;		
			}
			p++;
			if (p == 2) p = 0;
		}

		// Toggle LCD flag for printing only a temperature value every second
		var_LCD_flag = var_LCD_flag^0b00000001; 
		break;

	case STOP_ACQ_TIMER:

// Turn off the timer for counting sampling periods
		var_TMR_flag = 0;
		var_current_time = 0; 
		var_previous_time = 0; 	 

// For printing one time the IDLE message
		var_LCD_flag = 1;	

		break;
	}
}	


/*	===========================================================================
	ISR external interrupt INT0 pin 
This function is executed when an active edge is detected at pin PD2 (INT0)
The idea is simply set the global variable
	======================================================================== */
void interrupt_service_routine_INT1 (void){

	var_ST_flag = 1;  

	// LAB experimentation on debouncing techniques: 
	//delay(5); // wait some time 

}


/*	===========================================================================
	read_inputs 
	======================================================================== */
void read_inputs(void){

   if (digitalRead(pin_SW_HYS) == 1) // Active-low connected switch
		var_SW_HYS = 1;
   else
		var_SW_HYS = 0;
   
}

/*	===========================================================================
	write_outputs 
	======================================================================== */
void write_outputs(void){
   if (var_SLED == 1)
       digitalWrite(pin_SLED, LOW); 	// Active-low connected LED 
   else									
       digitalWrite(pin_SLED, HIGH);  


// PWM control voltage: // the power buffer is an inverter
	analogWrite(pin_Vc, 255 - var_PWM);

}


/*	===========================================================================
	Establishing the sampling frequency using internal timer
	======================================================================== */
void update_time_flag(void){

// Count real time only when acquiring
	if (var_TMR_flag == 1){
	  // From internal CLK: 
		var_current_time = millis();
		var_time_elapsed =  var_current_time - var_previous_time;  

		if(var_time_elapsed  > var_SP) {
				// is the flag is set is because time elapsed is sampling period
			var_SP_flag = 1;   
				// to start another sampling period cycle
			var_previous_time = var_current_time; 
		}
	}
}

